# Filename: controlhost.py
"""
A set of classes and tools wich uses the ControlHost protocol.

"""
import socket
import struct
import time

from .logger import get_logger

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)

BUFFER_SIZE = 1024


class Client(object):
    """The ControlHost client"""

    _valid_modes = ["any", "all"]

    def __init__(self, host, port=5553):
        self.host = host
        self.port = port
        self.socket = None
        self.tags = []
        self.valid_tags = []

    def subscribe(self, tag, mode="any"):
        if mode not in self._valid_modes:
            raise ValueError(
                "Possible subscription modes are: {}".format(
                    ", ".join(self._valid_modes)
                )
            )
        log.info("Subscribing to %s in mode %s", tag, mode)
        full_tag = self._full_tag(tag, mode)
        if full_tag not in self.tags:
            self.tags.append(full_tag)
        for t in tag.split():
            if t not in self.valid_tags:
                self.valid_tags.append(t)
        self._update_subscriptions()

    def unsubscribe(self, tag, mode="any"):
        try:
            self.tags.remove(self._full_tag(tag, mode))
            self.valid_tags.remove(tag)
        except ValueError:
            pass
        else:
            self._update_subscriptions()

    def _full_tag(self, tag, mode):
        mode_flag = " {} ".format("w" if mode == "any" else "a")
        full_tag = mode_flag + tag
        return full_tag

    def _update_subscriptions(self):
        log.debug("Subscribing to tags: %s", self.tags)
        if not self.socket:
            self._connect()
        tags = "".join(self.tags).encode("ascii")
        message = Message(b"_Subscri", tags)
        self.socket.send(message.data)
        message = Message(b"_Always")
        self.socket.send(message.data)

    def put_message(self, tag, data):
        """Send data to the ligier with a given tag"""
        if not self.socket:
            self._connect()
        msg = Message(tag, data)
        self.socket.send(msg.data)

    def get_message(self):
        while True:
            log.info("     Waiting for control host Prefix")
            if self.socket is None:
                log.error("Lost socket connection, reconnecting...")
                self._reconnect()
                continue
            try:
                data = self._recv(Prefix.SIZE)
                timestamp = time.time()
                log.info("    raw prefix data received: '%s'", data)
                if data == b"":
                    raise EOFError
                prefix = Prefix(data=data, timestamp=timestamp)
            except (UnicodeDecodeError, OSError, struct.error):
                log.error("Failed to construct Prefix, reconnecting.")
                self._reconnect()
                continue

            try:
                prefix_tag = str(prefix.tag)
            except UnicodeDecodeError:
                log.error("The tag could not be decoded. Reconnecting.")
                self._reconnect()
                continue

            if prefix_tag not in self.valid_tags:
                log.error(
                    "Invalid tag '%s' received, ignoring the message \n"
                    "and reconnecting.\n"
                    "  -> valid tags are: %s",
                    prefix_tag,
                    self.valid_tags,
                )
                self._reconnect()
                continue
            else:
                break

        log.info("       got a Prefix with %d bytes.", prefix.length)
        message = self._recv(prefix.length)
        log.info("     ------ returning message with %d bytes", len(message))
        return prefix, message

    def _connect(self):
        """Connect to JLigier"""
        log.debug("Connecting to JLigier")

        s = socket.socket()
        s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        s.connect((self.host, self.port))
        self.socket = s

    def _disconnect(self):
        """Close the socket"""
        log.debug("Disconnecting from JLigier")
        if self.socket:
            self.socket.close()

    def _reconnect(self):
        """Reconnect to JLigier and subscribe to the tags."""
        log.debug("Reconnecting to JLigier...")
        self._disconnect()
        self._connect()
        self._update_subscriptions()

    def _recv(self, size):
        """Receive the exact amount of bytes from the socket.

        This is needed since socket.recv(size) may return less than
        size, according to the specification: 'size' is the maximum
        number of bytes returned.
        """
        message = b""
        while len(message) < size:
            buffer_size = min((BUFFER_SIZE, (size - len(message))))
            try:
                if recv := self.socket.recv(buffer_size):
                    message += recv
                else:
                    log.error(
                        "Client has disconnected. "
                        + "Trying to reconnect in 30 seconds."
                    )
                    time.sleep(30)
                    self._reconnect()
                    message = b""
            except OSError as e:
                log.error("Failed to receive controlhost message: %s", e)
                raise BufferError
        return message

    def __enter__(self):
        self._connect()
        return self

    def __exit__(self, exit_type, value, traceback):
        self._disconnect()


class Message(object):
    """The representation of a ControlHost message."""

    def __init__(self, tag, message=b""):
        try:
            message = message.encode()
        except AttributeError:
            pass
        try:
            tag = tag.encode()
        except AttributeError:
            pass
        self.prefix = Prefix(tag, len(message))
        self.message = message

    @property
    def data(self):
        return self.prefix.data + self.message


class Tag(object):
    """Represents the tag in a ControlHost Prefix."""

    SIZE = 8

    def __init__(self, data=None):
        self._data = b""
        self.data = data

    @property
    def data(self):
        """The byte data"""
        return self._data

    @data.setter
    def data(self, value):
        """Set the byte data and fill up the bytes to fit the size."""
        if not value:
            value = b""
        if len(value) > self.SIZE:
            raise ValueError("The maximum tag size is {}".format(self.SIZE))
        self._data = value
        while len(self._data) < self.SIZE:
            self._data += b"\x00"

    def __str__(self):
        return self.data.decode(encoding="UTF-8").strip("\x00")

    def __len__(self):
        return len(self._data)


class Prefix(object):
    """The prefix of a ControlHost message."""

    SIZE = 16

    def __init__(self, tag=None, length=None, data=None, timestamp=None):
        if data:
            self.data = data
        else:
            self.tag = Tag(tag)
            self.length = length
        if timestamp is None:
            self.timestamp = time.time()
        else:
            self.timestamp = timestamp

    @property
    def data(self):
        return self.tag.data + struct.pack(">i", self.length) + b"\x00" * 4

    @data.setter
    def data(self, value):
        self.tag = Tag(data=value[: Tag.SIZE])
        self.length = struct.unpack(">i", value[Tag.SIZE : Tag.SIZE + 4])[0]

    def __str__(self):
        return "ControlHost Prefix with tag '{0}' ({1} bytes of data)".format(
            self.tag, self.length
        )



# Filename: test_mc.py
# pylint: disable=locally-disabled,C0111,R0904,C0103
import numpy as np

from km3pipe import Table, Blob
from km3pipe.testing import TestCase
from km3pipe.mc import geant2pdg, pdg2name, convert_mc_times_to_jte_times

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestMc(TestCase):
    def test_geant2pdg(self):
        self.assertEqual(22, geant2pdg(1))
        self.assertEqual(-13, geant2pdg(5))

    def test_geant2pdg_returns_0_for_unknown_particle_id(self):
        self.assertEqual(0, geant2pdg(-999))

    def test_pdg2name(self):
        self.assertEqual("mu-", pdg2name(13))
        self.assertEqual("nu(tau)~", pdg2name(-16))

    def test_pdg2name_returns_NA_for_unknown_particle(self):
        self.assertEqual("N/A", pdg2name(0))


class TestMCConvert(TestCase):
    def setUp(self):
        self.event_info = Table(
            {
                "timestamp": 1,
                "nanoseconds": 700000000,
                "mc_time": 1.74999978e9,
            }
        )

        self.mc_tracks = Table(
            {
                "time": 1,
            }
        )

        self.mc_hits = Table(
            {
                "time": 30.79,
            }
        )

        self.blob = Blob(
            {
                "event_info": self.event_info,
                "mc_hits": self.mc_hits,
                "mc_tracks": self.mc_tracks,
            }
        )

    def test_convert_mc_times_to_jte_times(self):
        times_mc_tracks = convert_mc_times_to_jte_times(
            self.mc_tracks.time,
            self.event_info.timestamp * 1e9 + self.event_info.nanoseconds,
            self.event_info.mc_time,
        )
        times_mc_hits = convert_mc_times_to_jte_times(
            self.mc_hits.time,
            self.event_info.timestamp * 1e9 + self.event_info.nanoseconds,
            self.event_info.mc_time,
        )

        assert times_mc_tracks is not None
        assert times_mc_hits is not None
        print(times_mc_tracks, times_mc_hits)
        assert np.allclose(times_mc_tracks, 49999781)
        assert np.allclose(times_mc_hits, 49999810.79)



#!/usr/bin/env python
# vim: ts=4 sw=4 et

from datetime import datetime
from collections import deque
import sys
from io import StringIO
import os
import shutil
import time
import threading

import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')    # noqa
import matplotlib.pyplot as plt
import matplotlib.dates as md
import pandas as pd

from km3pipe import Pipeline, Module
from km3pipe.config import Config
from km3pipe.io import CHPump
from km3pipe.io.daq import (DAQPreamble, DAQEvent)
from km3pipe.logger import get_logger
import km3pipe.style

km3pipe.style.use('km3pipe')
log = get_logger("trigger_rate")

PLOTS_PATH = 'www/plots'

xfmt = md.DateFormatter('%Y-%m-%d %H:%M')
lock = threading.Lock()


def trigger_rate_sampling_period():
    try:
        return int(Config().get("Monitoring", "trigger_rate_sampling_period"))
    except (TypeError, ValueError):
        return 180


class TriggerRate(Module):
    def configure(self):
        self.run = True
        self.interval = self.get("interval") or trigger_rate_sampling_period()
        self.event_times = deque(maxlen=4000)    # max events per interval
        # minutes to monitor
        self.trigger_rates = deque(maxlen=60 * 24 // (self.interval // 60))
        self.thread = threading.Thread(target=self.plot).start()
        self.store = pd.HDFStore('data/trigger_rates.h5')
        self.restore_data()

    def restore_data(self):
        with lock:
            try:
                data = zip(
                    self.store.trigger_rates.timestamp,
                    self.store.trigger_rates.rate
                )
                self.trigger_rates.extend(data)
                print(
                    "\n{0} data points restored.".format(
                        len(self.trigger_rates)
                    )
                )
            except AttributeError:
                pass

    def process(self, blob):
        if not str(blob['CHPrefix'].tag) == 'IO_EVT':
            return blob
        sys.stdout.write('.')
        sys.stdout.flush()

        data = blob['CHData']
        data_io = StringIO(data)
        preamble = DAQPreamble(file_obj=data_io)    # noqa
        event = DAQEvent(file_obj=data_io)
        timestamp = event.header.time_stamp
        with lock:
            self.event_times.append(timestamp)

        return blob

    def plot(self):
        while self.run:
            self.interval = trigger_rate_sampling_period()
            time.sleep(self.interval)
            with lock:
                self.create_plot()

    def create_plot(self):
        print('\n' + self.__class__.__name__ + ": updating plot.")
        timestamp = time.time()
        now = datetime.utcnow()
        interval = self.interval
        n_events = sum(t > timestamp - interval for t in self.event_times)
        rate = n_events / interval
        self.trigger_rates.append((now, rate))
        try:
            self.store.append(
                'trigger_rates',
                pd.DataFrame({
                    'timestamp': [now],
                    'rate': [rate]
                })
            )
        except pd.io.pytables.ClosedFileError:
            pass

        x, y = zip(*self.trigger_rates)
        if not any(y):
            return
        fig, ax = plt.subplots(figsize=(16, 4))
        ax.xaxis.set_major_formatter(xfmt)
        data = pd.DataFrame({'dates': x, 'rates': y})
        data.plot('dates', 'rates', grid=True, ax=ax, legend=False, style='.')
        ax.set_title(
            "Trigger Rate - via Event Times\n{0} UTC".format(
                datetime.utcnow().strftime("%c")
            )
        )
        ax.set_xlabel("time")
        ax.set_ylabel("trigger rate [Hz]")
        try:
            ax.set_yscale('log')
        except ValueError:
            pass

        fig.tight_layout()

        filename = os.path.join(PLOTS_PATH, 'trigger_rates.png')
        filename_tmp = os.path.join(PLOTS_PATH, 'trigger_rates_tmp.png')
        plt.savefig(filename_tmp, dpi=120, bbox_inches="tight")
        plt.close('all')
        shutil.move(filename_tmp, filename)
        print("Plot updated.")

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()
        if self.store.is_open:
            self.store.close()


pipe = Pipeline()
pipe.attach(CHPump, tags='IO_EVT', timeout=60 * 60 * 24 * 7, max_queue=2000)
pipe.attach(TriggerRate)
pipe.drain()



#!/usr/bin/env python3

from thepipe import Module, Blob
from ..dataclasses import Table
from .hdf5 import HDF5Header
from thepipe import Provenance

import km3io
import numpy as np
from collections import defaultdict


class OfflinePump(Module):
    def configure(self):
        self._filename = self.get("filename")
        step_size = self.get("step_size", default=2000)

        self._reader = km3io.OfflineReader(self._filename, step_size=step_size)
        self.header = self._reader.header
        self.blobs = self._blob_generator()

        Provenance().record_input(
            self._filename, uuid=str(self._reader.uuid), comment="OfflinePump input"
        )

        self.expose(self.header, "offline_header")

    def process(self, blob=None):
        return next(self.blobs)

    def finish(self):
        self._reader.close()

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.blobs)

    def __getitem__(self, item):
        if not isinstance(item, int):
            raise TypeError("Only integer indices are supported.")
        return Blob({"event": self._reader[item], "header": self.header})

    def get_number_of_blobs(self):
        return len(self._reader)

    def _blob_generator(self):
        for event in self._reader:
            blob = Blob({"event": event, "header": self.header})
            yield blob



# Filename: test_dataclasses.py
# pylint: disable=C0111,R0904,C0103
# vim:set ts=4 sts=4 sw=4 et:
"""
...

"""

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose
import pytest

import km3pipe.extras
from km3pipe.testing import TestCase, skip  # noqa
from km3pipe.dataclasses import (
    Table,
    NDArray,
    Vec3,
    inflate_dtype,
    has_structured_dt,
    is_structured,
    DEFAULT_H5LOC,
    DEFAULT_NAME,
    DEFAULT_SPLIT,
)

__author__ = "Tamas Gal, Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal, Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestDtypes(TestCase):
    def setUp(self):
        self.c_dt = np.dtype(
            [
                ("a", "<f4"),
                ("origin", "<u4"),
                ("pmt_id", "<u4"),
                ("time", "<f8"),
                ("group_id", "<u4"),
            ]
        )

    def test_is_structured(self):
        assert is_structured(self.c_dt)
        assert not is_structured(np.dtype("int64"))
        assert not is_structured(np.dtype(int))
        assert not is_structured(np.dtype(float))

    def test_has_structured_dt(self):
        assert has_structured_dt(np.ones(2, dtype=self.c_dt))
        assert not has_structured_dt(np.ones(2, dtype=float))
        assert not has_structured_dt(np.ones(2, dtype=int))
        assert not has_structured_dt([1, 2, 3])
        assert not has_structured_dt([1.0, 2.0, 3.0])
        assert not has_structured_dt([1.0, 2, 3.0])
        assert not has_structured_dt([])

    def test_inflate_hasstructured(self):
        arr = np.ones(3, dtype=self.c_dt)
        names = ["a", "b", "c"]
        print(arr.dtype)
        assert has_structured_dt(arr)
        dt_a = inflate_dtype(arr, names=names)
        assert dt_a == self.c_dt

    def test_inflate_nostructured(self):
        names = ["a", "b", "c"]
        arr = [1, 2, 3]
        assert not has_structured_dt(arr)
        dt_l = inflate_dtype(arr, names=names)
        assert dt_l == np.dtype([("a", "<i8"), ("b", "<i8"), ("c", "<i8")])

    def test_inflate_mixed_casts_up(self):
        arr = [1, 2, 3.0]
        names = ["a", "b", "c"]
        assert not has_structured_dt(arr)
        dt_a = inflate_dtype(arr, names=names)
        assert dt_a == np.dtype([("a", "<f8"), ("b", "<f8"), ("c", "<f8")])


class TestTable(TestCase):
    def setUp(self):
        self.dt = np.dtype([("a", int), ("b", float), ("group_id", int)])
        self.arr = np.array(
            [
                (0, 1.0, 2),
                (3, 7.0, 5),
                (6, 4.0, 8),
            ],
            dtype=self.dt,
        )

    def test_h5loc(self):
        tab = self.arr.view(Table)
        assert tab.h5loc == DEFAULT_H5LOC
        tab = Table(self.arr)
        assert tab.h5loc == DEFAULT_H5LOC
        tab = Table(self.arr, h5loc="/foo")
        assert tab.h5loc == "/foo"

    def test_split(self):
        tab = self.arr.view(Table)
        assert tab.split_h5 is False
        tab = Table(self.arr)
        assert tab.split_h5 is False
        tab = Table(self.arr, split_h5=True)
        assert tab.split_h5

    def test_name(self):
        tab = self.arr.view(Table)
        assert tab.name == DEFAULT_NAME
        tab = Table(self.arr)
        assert tab.name == DEFAULT_NAME
        tab = Table(self.arr, name="foo")
        assert tab.name == "foo"

    def test_view(self):
        tab = self.arr.view(Table)
        assert tab.dtype == self.dt
        assert tab.h5loc == DEFAULT_H5LOC
        assert_array_equal(tab.a, np.array([0, 3, 6]))
        assert tab[0]["group_id"] == 2
        assert tab[0].group_id == 2
        assert tab["group_id"][0] == 2
        assert tab.group_id[0] == 2
        assert isinstance(tab[0], np.record)
        for row in tab:
            assert isinstance(row, np.record)
            assert row["a"] == 0
            assert row.a == 0
            for c in row:
                assert c == 0
                break
            assert_allclose([0, 1.0, 2], [c for c in row])
            break

    def test_init(self):
        tab = Table(self.arr)
        assert tab.h5loc == DEFAULT_H5LOC
        tab = Table(self.arr, h5loc="/bla")
        assert tab.dtype == self.dt
        assert tab.h5loc == "/bla"
        assert_array_equal(tab.a, np.array([0, 3, 6]))
        assert tab[0]["group_id"] == 2
        assert tab[0].group_id == 2
        assert tab["group_id"][0] == 2
        assert tab.group_id[0] == 2
        assert isinstance(tab[0], np.record)
        for row in tab:
            assert isinstance(row, np.record)
            assert row["a"] == 0
            assert row.a == 0
            for c in row:
                assert c == 0
                break
            assert_allclose([0, 1.0, 2], [c for c in row])
            break

    def test_fromdict(self):
        n = 5
        dmap = {
            "a": np.ones(n, dtype=int),
            "b": np.zeros(n, dtype=float),
            "c": 0,
        }
        # tab = Table.from_dict(dmap)
        # self.assertTrue(isinstance(tab, Table))
        # assert tab.h5loc == DEFAULT_H5LOC
        dt = [("a", float), ("b", float), ("c", float)]
        tab = Table.from_dict(dmap, dtype=dt)
        assert tab.h5loc == DEFAULT_H5LOC
        assert isinstance(tab, Table)
        tab = Table.from_dict(dmap, dtype=dt, h5loc="/foo")
        assert tab.h5loc == "/foo"
        assert isinstance(tab, Table)
        bad_dt = [("a", float), ("b", float), ("c", float), ("d", int)]
        with pytest.raises(KeyError):
            tab = Table.from_dict(dmap, dtype=bad_dt)

    def test_from_dict_without_dtype(self):
        data = {"b": [1, 2], "c": [3, 4], "a": [5, 6]}
        tab = Table.from_dict(data)
        assert np.allclose([1, 2], tab.b)
        assert np.allclose([3, 4], tab.c)
        assert np.allclose([5, 6], tab.a)

    def test_from_dict_with_unordered_columns_wrt_to_dtype_fields(self):
        data = {"b": [1, 2], "c": [3, 4], "a": [5, 6]}
        dt = [("a", float), ("b", float), ("c", float)]
        tab = Table.from_dict(data, dtype=dt)
        assert np.allclose([1, 2], tab.b)
        assert np.allclose([3, 4], tab.c)
        assert np.allclose([5, 6], tab.a)

    def test_from_dict_with_fillna(self):
        data = {"a": [1, 2], "b": [3, 4]}
        dt = [("a", float), ("b", float), ("c", float)]
        tab = Table.from_dict(data, dtype=dt, fillna=True)
        assert np.isnan(tab.c[0])
        assert np.isnan(tab.c[1])

    def test_init_implicitly_from_dict_with_fillna(self):
        data = {"a": [1, 2], "b": [3, 4]}
        dt = [("a", float), ("b", float), ("c", float)]
        tab = Table(data, dtype=dt, fillna=True)
        assert np.isnan(tab.c[0])
        assert np.isnan(tab.c[1])

    def test_from_dict_doesnt_alter_original_dict(self):
        a = [1, 2]
        b = [False, True]
        c = 3
        data = {"a": a, "b": b, "c": c}
        tab = Table(data)
        assert data["a"] is a
        assert data["b"] is b
        assert data["c"] is c

    def test_fromcolumns(self):
        n = 5
        dlist = [
            np.ones(n, dtype=int),
            np.zeros(n, dtype=float),
            0,
        ]
        dt = np.dtype([("a", float), ("b", float), ("c", float)])
        with pytest.raises(ValueError):
            tab = Table(dlist, dtype=dt)
        tab = Table.from_columns(dlist, dtype=dt)
        print(tab.dtype)
        print(tab.shape)
        print(tab)
        assert tab.h5loc == DEFAULT_H5LOC
        assert isinstance(tab, Table)
        tab = Table.from_columns(dlist, dtype=dt, h5loc="/foo")
        print(tab.dtype)
        print(tab.shape)
        print(tab)
        assert tab.h5loc == "/foo"
        assert isinstance(tab, Table)
        bad_dt = [("a", float), ("b", float), ("c", float), ("d", int)]
        with pytest.raises(ValueError):
            tab = Table.from_columns(dlist, dtype=bad_dt)
            print(tab.dtype)
            print(tab.shape)
            print(tab)

    def test_from_columns_with_colnames(self):
        t = Table.from_columns(
            [
                [1, 2, 3],
                [4, 5, 6],
                [7, 8, 9],
                [10, 11, 12],
                [13, 14, 15],
                [16, 17, 18],
                [19, 20, 21],
            ],
            colnames=["a", "b", "c", "d", "e", "f", "g"],
        )
        print("t.a: {}".format(t.a))
        assert np.allclose([1, 2, 3], t.a)
        print("t.b: {}".format(t.b))
        assert np.allclose([4, 5, 6], t.b)

    def test_from_columns_with_colnames_upcasts(self):
        t = Table.from_columns([[1, 2, 3], [4, 5.0, 6]], colnames=["a", "b"])
        assert t.dtype == np.dtype([("a", float), ("b", float)])

    def test_from_columns_with_mismatching_columns_and_dtypes_raises(self):
        with pytest.raises(ValueError):
            Table.from_columns([[1, 2, 3], [4, 5, 6]], dtype=np.dtype([("a", "f4")]))

    def test_from_rows_with_colnames(self):
        t = Table.from_rows([[1, 2], [3, 4], [5, 6]], colnames=["a", "b"])
        assert t.dtype == np.dtype([("a", int), ("b", int)])
        assert np.allclose([1, 3, 5], t.a)
        assert np.allclose([2, 4, 6], t.b)

    def test_from_rows_with_colnames_upcasts(self):
        t = Table.from_rows([[1, 2], [3.0, 4], [5, 6]], colnames=["a", "b"])
        assert t.dtype == np.dtype([("a", float), ("b", float)])

    def test_from_rows_dim(self):
        t = Table.from_rows([[1, 2], [3.0, 4], [5, 6]], colnames=["a", "b"])
        assert t.shape == (3,)

    def test_from_columns_dim(self):
        t = Table.from_columns([[1, 2, 3], [4, 5.0, 6]], colnames=["a", "b"])
        assert t.shape == (3,)

    def test_fromrows(self):
        dlist = [
            [1, 2, 3],
            [4, 5, 6],
        ]
        dt = np.dtype([("a", float), ("b", float), ("c", float)])
        with pytest.raises(ValueError):
            tab = Table(dlist, dtype=dt)
        tab = Table.from_rows(dlist, dtype=dt)
        print(tab.dtype)
        print(tab.shape)
        print(tab)
        assert tab.h5loc == DEFAULT_H5LOC
        assert isinstance(tab, Table)
        tab = Table.from_rows(dlist, dtype=dt, h5loc="/foo")
        print(tab.dtype)
        print(tab.shape)
        print(tab)
        assert tab.h5loc == "/foo"
        assert isinstance(tab, Table)
        bad_dt = [("a", float), ("b", float), ("c", float), ("d", int)]
        with pytest.raises(ValueError):
            tab = Table.from_rows(dlist, dtype=bad_dt)
            print(tab.dtype)
            print(tab.shape)
            print(tab)

    def test_expand_scalars(self):
        dmap = {
            "a": 1,
            "b": 0.0,
            "c": 0,
        }
        t1 = Table._expand_scalars(dmap)
        assert len(t1) > 0
        dmap2 = {
            "a": [1, 2, 1],
            "b": 0.0,
            "c": [0, 1],
        }
        t2 = Table._expand_scalars(dmap2)
        assert len(t2) > 0
        dmap3 = {
            "a": [1, 2, 1],
            "b": [0.0],
            "c": [0, 1],
        }
        t3 = Table._expand_scalars(dmap3)
        assert len(t3) > 0
        dmap4 = {
            "a": [1, 2, 1],
            "b": np.array(0.0),
            "c": [0, 1],
        }
        t4 = Table._expand_scalars(dmap4)
        assert len(t4) > 0
        dmap5 = {
            "a": [1, 2, 1],
            "b": np.array([1]),
            "c": [0, 1],
        }
        t5 = Table._expand_scalars(dmap5)
        assert len(t5) > 0

    def test_from_flat_dict(self):
        dmap = {
            "a": 1,
            "b": 0.0,
            "c": 0,
        }
        # tab = Table.from_dict(dmap)
        # self.assertTrue(isinstance(tab, Table))
        # assert tab.h5loc == DEFAULT_H5LOC
        dt = [("a", float), ("b", float), ("c", float)]
        tab = Table.from_dict(dmap, dtype=dt)
        assert tab.h5loc == DEFAULT_H5LOC
        assert isinstance(tab, Table)
        tab = Table.from_dict(dmap, dtype=dt, h5loc="/foo")
        assert tab.h5loc == "/foo"
        assert isinstance(tab, Table)
        bad_dt = [("a", float), ("b", float), ("c", float), ("d", int)]
        with pytest.raises(KeyError):
            tab = Table.from_dict(dmap, dtype=bad_dt)

    def test_from_mixed_dict(self):
        dmap = {
            "a": 1,
            "b": 0.0,
            "c": np.zeros(4),
        }
        # tab = Table.from_dict(dmap)
        # self.assertTrue(isinstance(tab, Table))
        # assert tab.h5loc == DEFAULT_H5LOC
        dt = [("a", float), ("b", float), ("c", float)]
        tab = Table.from_dict(dmap, dtype=dt)
        assert tab.h5loc == DEFAULT_H5LOC
        assert isinstance(tab, Table)
        tab = Table.from_dict(dmap, dtype=dt, h5loc="/foo")
        assert tab.h5loc == "/foo"
        assert isinstance(tab, Table)
        bad_dt = [("a", float), ("b", float), ("c", float), ("d", int)]
        with pytest.raises(KeyError):
            tab = Table.from_dict(dmap, dtype=bad_dt)

    def test_from_2d(self):
        l2d = [(0, 1), (2, 3), (4, 5), (6, 7), (8, 9)]
        names = ["a", "origin", "pmt_id", "time", "group_id"]
        dta = inflate_dtype(l2d, names)
        with pytest.raises(ValueError):
            t = Table(l2d)
        with pytest.raises(ValueError):
            t = Table(l2d, dtype=None)
        with pytest.raises(ValueError):
            t = Table(l2d, colnames=names)
        with pytest.raises(ValueError):
            t = Table(l2d, dtype=dta)
        with pytest.raises(ValueError):
            t = Table(l2d, dtype=dta, colnames=["a", "b", "c", "d"])  # noqa

    def test_flat_raises(self):
        with pytest.raises(ValueError):
            t = Table([1, 2, 3], dtype=int).dtype
        with pytest.raises(ValueError):
            t = Table([1, 2, 3], dtype=float).dtype
        with pytest.raises(ValueError):
            t = Table([1, 2, 3], dtype=None).dtype
        with pytest.raises(ValueError):
            t = Table([1, 2, 3]).dtype
        with pytest.raises(ValueError):
            t = Table([1, 2, 3], colnames=["a", "b", "c"])  # noqa

    def test_init_with_unstructured_raises_valueerror(self):
        with pytest.raises(ValueError):
            Table(np.array([[1, 2, 3], [4, 5, 6]]))

    def test_fromdict_init(self):
        n = 5
        dmap = {
            "a": np.ones(n, dtype=int),
            "b": np.zeros(n, dtype=float),
            "c": 0,
        }
        dt = [("a", float), ("b", float), ("c", float)]
        tab = Table(dmap, dtype=dt)
        assert tab.h5loc == DEFAULT_H5LOC
        self.assertTrue(isinstance(tab, Table))
        tab = Table(dmap, dtype=dt, h5loc="/foo")
        assert tab.h5loc == "/foo"
        self.assertTrue(isinstance(tab, Table))
        bad_dt = [("a", float), ("b", float), ("c", float), ("d", int)]
        with pytest.raises(KeyError):
            tab = Table(dmap, dtype=bad_dt)

    def test_from_record(self):
        t = Table({"a": [0, 1, 2], "b": [10, 20, 30]})
        t2 = Table(t[1])
        assert 1 == t2[0].a
        assert 1 == t2.a[0]
        assert 20 == t2[0].b
        assert 20 == t2.b[0]

    def test_append_columns(self):
        tab = Table(self.arr)
        print(tab)
        with pytest.raises(ValueError):
            tab = tab.append_columns("new", [1, 2, 3, 4])
        tab = tab.append_columns("new", [1, 2, 3])
        print(tab)
        assert tab.new[0] == 1
        assert tab.new[-1] == 3
        tab = tab.append_columns("bar", 0)
        print(tab)
        assert tab.bar[0] == 0
        assert tab.bar[-1] == 0
        tab = tab.append_columns("lala", [1])
        print(tab)
        assert tab.lala[0] == 1
        assert tab.lala[-1] == 1
        with pytest.raises(ValueError):
            tab = tab.append_columns(["m", "n"], [1, 2])
        with pytest.raises(ValueError):
            tab = tab.append_columns(["m", "n"], [[1], [2]])
        tab = tab.append_columns(["m", "n"], [[1, 1, 2], [2, 4, 5]])
        print(tab)
        assert tab.m[0] == 1
        assert tab.m[-1] == 2
        assert tab.n[0] == 2
        assert tab.n[-1] == 5

    def test_append__single_column(self):
        tab = Table({"a": 1})
        print(tab.dtype)
        tab = tab.append_columns(["b"], np.array([[2]]))
        print(tab.dtype)
        print(tab.b)

    def test_append_columns_with_single_value(self):
        tab = Table({"a": 1})
        tab = tab.append_columns("group_id", 0)
        assert 0 == tab.group_id[0]

    def test_append_columns_with_multiple_values(self):
        tab = Table({"a": [1, 2]})
        tab = tab.append_columns("group_id", [0, 1])
        assert 0 == tab.group_id[0]
        assert 1 == tab.group_id[1]

    def test_append_columns_modifies_dtype(self):
        tab = Table({"a": [1, 2]})
        tab = tab.append_columns("group_id", [0, 1])
        assert "group_id" in tab.dtype.names

    def test_append_column_which_is_too_short_raises(self):
        tab = Table({"a": [1, 2, 3]})
        with pytest.raises(ValueError):
            tab = tab.append_columns("b", [4, 5])

    def test_append_columns_duplicate(self):
        tab = Table({"a": 1})
        with pytest.raises(ValueError):
            tab = tab.append_columns(["a"], np.array([[2]]))

    def test_append_columns_with_mismatching_lengths_raises(self):
        tab = Table({"a": [1, 2, 3]})
        with pytest.raises(ValueError):
            tab.append_columns(colnames=["b", "c"], values=[[4, 5, 6], [7, 8]])

    def test_append_columns_which_is_too_long(self):
        tab = Table({"a": [1, 2, 3]})
        with pytest.raises(ValueError):
            tab.append_columns("b", values=[4, 5, 6, 7])

    def test_drop_column(self):
        tab = Table({"a": 1, "b": 2})
        tab = tab.drop_columns("a")
        with pytest.raises(AttributeError):
            print(tab.a)
        print(tab.b)

    def test_drop_columns(self):
        tab = Table({"a": 1, "b": 2, "c": 3})
        print(tab.dtype)
        tab = tab.drop_columns(["a", "b"])
        print(tab.dtype)
        with pytest.raises(AttributeError):
            print(tab.a)
        with pytest.raises(AttributeError):
            print(tab.b)
        print(tab.c)

    def test_template(self):
        n = 10
        channel_ids = np.arange(n)
        dom_ids = np.arange(n)
        times = np.arange(n)
        tots = np.arange(n)
        triggereds = np.ones(n)
        d_hits = {
            "channel_id": channel_ids,
            "dom_id": dom_ids,
            "time": times,
            "tot": tots,
            "triggered": triggereds,
            "group_id": 0,  # event_id
        }
        tab = Table.from_template(d_hits, "Hits")
        assert tab.name == "Hits"
        assert tab.split_h5 is True
        assert isinstance(tab, Table)
        ar_hits = {
            "channel_id": np.ones(n, dtype=int),
            "dom_id": np.ones(n, dtype=int),
            "time": np.ones(n, dtype=float),
            "tot": np.ones(n, dtype=float),
            "triggered": np.ones(n, dtype=bool),
            "group_id": np.ones(n, dtype=int),
        }
        tab = Table.from_template(ar_hits, "Hits")
        assert tab.name == "Hits"
        assert tab.split_h5 is True
        assert isinstance(tab, Table)

    def test_incomplete_template(self):
        n = 10
        channel_ids = np.arange(n)
        dom_ids = np.arange(n)
        # times = np.arange(n)
        tots = np.arange(n)
        triggereds = np.ones(n)
        d_hits = {
            "channel_id": channel_ids,
            "dom_id": dom_ids,
            # 'time': times,
            "tot": tots,
            "triggered": triggereds,
            "group_id": 0,  # event_id
        }
        with pytest.raises(KeyError):
            tab = Table.from_template(d_hits, "Hits")
            assert tab is not None
        ar_hits = {
            "channel_id": np.ones(n, dtype=int),
            "dom_id": np.ones(n, dtype=int),
            # 'time': np.ones(n, dtype=float),
            "tot": np.ones(n, dtype=float),
            "triggered": np.ones(n, dtype=bool),
            "group_id": np.ones(n, dtype=int),
        }
        with pytest.raises(KeyError):
            tab = Table.from_template(ar_hits, "Hits")
            assert tab is not None

    def test_adhoc_template(self):
        a_template = {
            "dtype": np.dtype([("a", "<u4"), ("b", "f4")]),
            "h5loc": "/yat",
            "split_h5": True,
            "h5singleton": True,
            "name": "YetAnotherTemplate",
        }
        arr = np.array([(1, 3), (2, 4)], dtype=a_template["dtype"])
        tab = Table.from_template(arr, a_template)
        self.assertListEqual([1, 2], list(tab.a))
        self.assertListEqual([3.0, 4.0], list(tab.b))
        assert "YetAnotherTemplate" == tab.name
        assert tab.h5singleton

    def test_adhoc_noname_template(self):
        a_template = {
            "dtype": np.dtype([("a", "<u4"), ("b", "f4")]),
            "h5loc": "/yat",
            "split_h5": True,
            "h5singleton": True,
        }
        arr = np.array([(1, 3), (2, 4)], dtype=a_template["dtype"])
        tab = Table.from_template(arr, a_template)
        self.assertListEqual([1, 2], list(tab.a))
        self.assertListEqual([3.0, 4.0], list(tab.b))
        assert DEFAULT_NAME == tab.name
        assert tab.h5singleton

    def test_element_list_with_dtype(self):
        bad_elist = [
            [1, 2.1],
            [3, 4.2],
            [5, 6.3],
        ]
        dt = np.dtype([("a", int), ("b", float)])
        print("list(list)")
        arr_bad = np.array(bad_elist, dtype=dt)
        print(arr_bad)
        elist = [tuple(el) for el in bad_elist]
        arr = np.array(elist, dtype=dt)
        print("list(tuple)")
        print(arr)
        tab = Table(arr)
        print(tab)
        assert tab.a[0] == 1

    def test_sort(self):
        dt = np.dtype([("a", int), ("b", float), ("c", int)])
        arr = np.array(
            [
                (0, 1.0, 2),
                (3, 7.0, 5),
                (6, 4.0, 8),
            ],
            dtype=dt,
        )
        tab = Table(arr)
        tab_sort = tab.sorted("b")
        assert_array_equal(tab_sort["a"], np.array([0, 6, 3]))

    def test_init_directly_with_df(self):
        pd = km3pipe.extras.pandas()

        df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        tab = Table(df, h5loc="/foo")
        assert np.allclose(df.a, tab.a)
        assert np.allclose(df.b, tab.b)
        assert tab.h5loc == "/foo"

    def test_df(self):
        pd = km3pipe.extras.pandas()

        dt = np.dtype([("a", int), ("b", float), ("c", int)])
        arr = np.array(
            [
                (0, 1.0, 2),
                (3, 7.0, 5),
                (6, 4.0, 8),
            ],
            dtype=dt,
        )
        print(dir(Table))
        df = pd.DataFrame(arr)
        tab = Table.from_dataframe(df, h5loc="/bla")
        df2 = tab.to_dataframe()
        pd.testing.assert_frame_equal(df, df2)

    def test_slicing(self):
        dt = np.dtype([("a", int), ("b", float), ("c", bool)])
        arr = np.array(
            [
                (0, 1.0, True),
                (2, 3.0, False),
                (4, 5.0, True),
            ],
            dtype=dt,
        )
        tab = Table(arr)
        assert 2 == len(tab[tab.c])
        assert 1 == len(tab[tab.b > 3.0])

    def test_contains(self):
        dt = np.dtype([("a", int), ("b", float), ("c", bool)])
        arr = np.array(
            [
                (0, 1.0, True),
                (2, 3.0, False),
                (4, 5.0, True),
            ],
            dtype=dt,
        )
        tab = Table(arr)
        assert "a" in tab
        assert "b" in tab
        assert "c" in tab
        assert "d" not in tab

    def test_index_returns_reference(self):
        tab = Table({"a": [1, 2, 3]})
        tab[1].a = 4
        assert np.allclose(tab.a, [1, 4, 3])

    def test_index_of_attribute_returns_reference(self):
        tab = Table({"a": [1, 2, 3]})
        tab.a[1] = 4
        assert np.allclose(tab.a, [1, 4, 3])

    def test_mask_returns_copy(self):
        tab = Table({"a": [1, 2, 3]})
        tab[[True, False, True]].a = [4, 5]
        assert np.allclose(tab.a, [1, 2, 3])

    def test_mask_on_attribute_returns_reference(self):
        tab = Table({"a": [1, 2, 3]})
        tab.a[[True, False, True]] = [4, 5]
        assert np.allclose(tab.a, [4, 2, 5])

    def test_index_mask_returns_copy(self):
        tab = Table({"a": [1, 2, 3]})
        tab[[1, 2]].a = [4, 5]
        assert np.allclose(tab.a, [1, 2, 3])

    def test_index_mask_of_attribute_returns_reference(self):
        tab = Table({"a": [1, 2, 3]})
        tab.a[[1, 2]] = [4, 5]
        assert np.allclose(tab.a, [1, 4, 5])

    def test_slice_returns_reference(self):
        tab = Table({"a": [1, 2, 3]})
        tab[:2].a = [4, 5]
        assert np.allclose(tab.a, [4, 5, 3])

    def test_slice_of_attribute_returns_reference(self):
        tab = Table({"a": [1, 2, 3]})
        tab.a[:2] = [4, 5]
        assert np.allclose(tab.a, [4, 5, 3])

    def test_slice_keeps_metadata(self):
        tab = Table(
            {"a": [1, 2, 3]},
            h5loc="/lala",
            split_h5=True,
            name="bla",
            h5singleton=False,
        )
        assert tab[:2].h5loc == "/lala"
        assert tab[:2].name == "bla"
        assert not tab[:2].h5singleton
        assert tab[:2].split_h5

    def test_mask_keeps_metadata(self):
        tab = Table(
            {"a": [1, 2, 3]}, h5loc="/lala", split_h5=True, name="bla", h5singleton=True
        )
        m = np.ones(len(tab), dtype=bool)
        assert tab[m].h5loc == "/lala"
        assert tab[m].name == "bla"
        assert tab[m].h5singleton
        assert tab[m].split_h5

    def test_indexing_keeps_metadata(self):
        tab = Table(
            {"a": [1, 2, 3]}, h5loc="/lala", split_h5=True, name="bla", h5singleton=True
        )
        im = [1, 1, 0]
        assert tab[im].h5loc == "/lala"
        assert tab[im].name == "bla"
        assert tab[im].h5singleton
        assert tab[im].split_h5

    def test_crash_repr(self):
        a = np.array("", dtype=[("a", "<U1")])
        with pytest.raises(TypeError):
            print(len(a))
        tab = Table(a)
        s = tab.__str__()
        assert s is not None
        r = tab.__repr__()
        assert r is not None

    def test_array_finalize_with_obj_none(self):
        tab = Table({"a": [1, 2, 3]})
        assert tab.__array_finalize__(None) is None

    def test_array_wrap(self):
        t = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        wrapped = t.__array_wrap__(np.array((Table({"a": 1}))))
        assert wrapped.a[0] == 1

    def test_templates_avail(self):
        t = Table({"a": 1})
        templates = t.templates_avail
        assert templates

    def test_add_table_to_itself(self):
        tab = Table({"a": [1]})
        added_tab = tab + tab
        assert 2 == len(added_tab)
        self.assertListEqual([1, 1], list(added_tab.a))

    def test_add_two_tables(self):
        tab1 = Table({"a": [1, 2]})
        tab2 = Table({"a": [3, 4]})
        added_tab = tab1 + tab2
        assert 4 == len(added_tab)
        self.assertListEqual([1, 2, 3, 4], list(added_tab.a))

    def test_add_two_tables_with_different_lengths(self):
        tab1 = Table({"a": [1, 2]})
        tab2 = Table({"a": [3, 4, 5]})
        added_tab = tab1 + tab2
        assert 5 == len(added_tab)
        self.assertListEqual([1, 2, 3, 4, 5], list(added_tab.a))

    def test_add_two_tables_with_different_lengths_and_columns(self):
        tab1 = Table({"a": [1, 2], "b": [100, 200]})
        tab2 = Table({"a": [3, 4, 5], "b": [300, 400, 500]})
        added_tab = tab1 + tab2
        assert 5 == len(added_tab)
        self.assertListEqual([1, 2, 3, 4, 5], list(added_tab.a))
        self.assertListEqual([100, 200, 300, 400, 500], list(added_tab.b))

    def test_adding_preserves_metadata(self):
        tab1 = Table(
            {"a": [1, 2]}, h5loc="/a", h5singleton=True, split_h5=True, name="FooTable"
        )
        tab2 = Table({"a": [3, 4, 5]})
        added_tab = tab1 + tab2
        assert "/a" == tab1.h5loc
        assert added_tab.h5singleton
        assert added_tab.split_h5
        assert "FooTable" == added_tab.name

    def test_add_tables_with_same_colnames_but_different_dtype_order(self):
        cols1 = ("b", "a")
        tab1 = Table.from_columns([[100, 200], [1, 2]], colnames=cols1)
        self.assertTupleEqual(cols1, tab1.dtype.names)
        cols2 = ("a", "b")
        tab2 = Table.from_columns([[3, 4, 5], [300, 400, 500]], colnames=cols2)
        added_tab = tab1 + tab2
        self.assertListEqual([1, 2, 3, 4, 5], list(added_tab.a))
        self.assertListEqual([100, 200, 300, 400, 500], list(added_tab.b))
        self.assertListEqual(list(added_tab.dtype.names), list(tab1.dtype.names))

    def test_add_table_with_different_cols(self):
        tab1 = Table({"a": [1]})
        tab2 = Table({"b": [2]})
        with self.assertRaises(TypeError):
            added_tab = tab1 + tab2

    def test_merge(self):
        tab1 = Table({"a": [1]}, h5loc="/a", h5singleton=True)
        tab2 = Table({"a": [2]})
        tab3 = Table({"a": [3]})
        merged_tab = Table.merge([tab1, tab2, tab3])
        assert 3 == len(merged_tab)
        self.assertListEqual([1, 2, 3], list(merged_tab.a))
        assert "/a" == merged_tab.h5loc
        assert merged_tab.h5singleton

    def test_merge_different_columns_with_no_nan_compatible_dtype(self):
        tab1 = Table({"a": [1]}, h5loc="/a", h5singleton=True)
        tab2 = Table({"b": [2]})
        tab3 = Table({"c": [3]})
        with self.assertRaises(ValueError):
            merged_tab = Table.merge([tab1, tab2, tab3])

    def test_merge_different_columns_with_no_nan_compatible_dtype_even_if_fillna(self):
        tab1 = Table({"a": [1]}, h5loc="/a", h5singleton=True)
        tab2 = Table({"b": [2]})
        tab3 = Table({"c": [3]})
        with self.assertRaises(ValueError):
            merged_tab = Table.merge([tab1, tab2, tab3], fillna=True)

    def test_merge_different_columns_fills_nan_when_fillna(self):
        tab1 = Table({"a": [1.1]}, h5loc="/a", h5singleton=True, split_h5=True)
        tab2 = Table({"b": [2.2]})
        tab3 = Table({"c": [3.3]})
        merged_tab = Table.merge([tab1, tab2, tab3], fillna=True)
        assert 3 == len(merged_tab)

        assert 1.1 == merged_tab.a[0]
        assert np.isnan(merged_tab.a[1])
        assert np.isnan(merged_tab.a[2])

        assert np.isnan(merged_tab.b[0])
        assert 2.2 == merged_tab.b[1]
        assert np.isnan(merged_tab.b[2])

        assert np.isnan(merged_tab.c[0])
        assert np.isnan(merged_tab.c[1])
        assert 3.3 == merged_tab.c[2]

        assert "/a" == merged_tab.h5loc
        assert merged_tab.h5singleton
        assert merged_tab.split_h5

    def test_merge_other_different_columns_fills_nan_when_fillna(self):
        tab1 = Table({"a": [1.1, 1.2], "b": [10.1, 10.2]})
        tab2 = Table({"a": [2.1, 2.2], "c": [100.1, 100.2]})

        merged_tab = Table.merge([tab1, tab2], fillna=True)

        assert 4 == len(merged_tab)

        self.assertListEqual([1.1, 1.2, 2.1, 2.2], list(merged_tab.a))
        self.assertListEqual([10.1, 10.2], list(merged_tab.b[:2]))
        self.assertListEqual([100.1, 100.2], list(merged_tab.c[2:]))
        assert np.isnan(merged_tab.c[0])
        assert np.isnan(merged_tab.c[1])
        assert np.isnan(merged_tab.b[2])
        assert np.isnan(merged_tab.b[3])

    def test_init_with_different_dicts_but_same_content(self):
        t1 = Table({"a": [1, 2], "b": [3, 4], "c": [5, 6]})
        t2 = Table({"c": [1, 2], "a": [3, 4], "b": [5, 6]})

        assert t1.dtype == t2.dtype

    def test_init_from_template_with_differently_ordered_dicts(self):
        t1 = Table.from_template(
            {
                "frame_index": 1,
                "slice_id": 2,
                "timestamp": 3,
                "nanoseconds": 4,
                "n_frames": 5,
            },
            "TimesliceInfo",
        )
        t2 = Table.from_template(
            {
                "n_frames": 5,
                "timestamp": 3,
                "nanoseconds": 4,
                "slice_id": 2,
                "frame_index": 1,
            },
            "TimesliceInfo",
        )
        assert t1.dtype == t2.dtype
        assert t1.frame_index[0] == t2.frame_index[0]
        assert t1.slice_id[0] == t2.slice_id[0]
        assert t1.nanoseconds[0] == t2.nanoseconds[0]
        assert t1.n_frames[0] == t2.n_frames[0]
        assert t1.timestamp[0] == t2.timestamp[0]

    def test_reorder_dtypes(self):
        dtype = np.dtype([("a", "<i8"), ("c", "<i8"), ("b", "<f8")])
        dtype_reordered = np.dtype([("b", "<f8"), ("c", "<i8"), ("a", "<i8")])

        tab = Table({"a": 1, "b": 2.5, "c": 3}, dtype=dtype)
        assert tab.dtype == dtype
        assert 1 == tab.a[0]
        assert 2.5 == tab.b[0]
        assert 3 == tab.c[0]

        tab_reordered = Table(tab, dtype=dtype_reordered)
        assert tab_reordered.dtype == dtype_reordered
        assert 1 == tab_reordered.a[0]
        assert 2.5 == tab_reordered.b[0]
        assert 3 == tab_reordered.c[0]

    def test_reorder_dtypes_with_differing_names_raises(self):
        dtype = np.dtype([("a", "<i8"), ("c", "<i8"), ("b", "<f8")])
        dtype_reordered = np.dtype([("b", "<f8"), ("a", "<i8")])
        tab = Table({"a": 1, "b": 2.5, "c": 3}, dtype=dtype)

        with self.assertRaises(ValueError):
            tab2 = Table(tab, dtype=dtype_reordered)

    def test_reorder_dtypes_w_matching_names_but_different_types_raise(self):
        dtype = np.dtype([("a", "<i8"), ("c", "<i8"), ("b", "<f8")])
        dtype_reordered = np.dtype([("a", "<f8"), ("c", "<i8"), ("b", "<f8")])
        tab = Table({"a": 1, "b": 2.5, "c": 3}, dtype=dtype)

        with self.assertRaises(ValueError):
            tab2 = Table(tab, dtype=dtype_reordered)


class TestTableFancaAttributes(TestCase):
    def setUp(self):
        self.arr_bare = Table(
            {
                "a": [1, 2, 3],
                "b": [3, 4, 5],
            }
        )
        self.arr_wpos = Table(
            {
                "a": [1, 2, 3],
                "b": [3, 4, 5],
                "pos_x": [10, 20, 30],
                "pos_y": [40, 50, 60],
                "pos_z": [70, 80, 90],
                "dir_x": [10.0, 20.0, 30.0],
                "dir_y": [40.0, 50.0, 60.0],
                "dir_z": [70.0, 80.0, 90.0],
            }
        )

    def test_pos_getter(self):
        tab = Table({"pos_x": [1, 2, 3], "pos_y": [4, 5, 6], "pos_z": [7, 8, 9]})
        assert np.allclose([[1, 4, 7], [2, 5, 8], [3, 6, 9]], tab.pos)

    def test_pos_getter_for_single_entry(self):
        tab = Table({"pos_x": [1, 2, 3], "pos_y": [4, 5, 6], "pos_z": [7, 8, 9]})
        assert np.allclose([[2, 5, 8]], tab.pos[1])

    def test_dir_getter(self):
        tab = Table({"dir_x": [1, 2, 3], "dir_y": [4, 5, 6], "dir_z": [7, 8, 9]})
        assert np.allclose([[1, 4, 7], [2, 5, 8], [3, 6, 9]], tab.dir)

    def test_dir_getter_for_single_entry(self):
        tab = Table({"dir_x": [1, 2, 3], "dir_y": [4, 5, 6], "dir_z": [7, 8, 9]})
        assert np.allclose([[2, 5, 8]], tab.dir[1])

    def test_dir_setter(self):
        tab = Table({"dir_x": [1, 0, 0], "dir_y": [0, 1, 0], "dir_z": [0, 0, 1]})
        new_dir = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tab.dir = new_dir
        assert np.allclose(new_dir, tab.dir)

    def test_pos_setter(self):
        tab = Table({"pos_x": [1, 0, 0], "pos_y": [0, 1, 0], "pos_z": [0, 0, 1]})
        new_pos = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        tab.pos = new_pos
        assert np.allclose(new_pos, tab.pos)

    def test_phi(self):
        tab = Table({"dir_x": [1, 0, 0], "dir_y": [0, 1, 0], "dir_z": [0, 0, 1]})
        p = tab.phi
        assert p is not None

    def test_phi(self):
        tab = Table({"dir_x": [1, 0, 0], "dir_y": [0, 1, 0], "dir_z": [0, 0, 1]})
        p = tab.theta
        assert p is not None

    def test_zen(self):
        tab = Table({"dir_x": [1, 0, 0], "dir_y": [0, 1, 0], "dir_z": [0, 0, 1]})
        p = tab.zenith
        assert p is not None

    def test_azi(self):
        tab = Table({"dir_x": [1, 0, 0], "dir_y": [0, 1, 0], "dir_z": [0, 0, 1]})
        p = tab.azimuth
        assert p is not None

    def test_triggered_keeps_attrs(self):
        n = 5
        channel_ids = np.arange(n)
        dom_ids = np.arange(n)
        times = np.arange(n)
        tots = np.arange(n)
        triggereds = np.array([0, 1, 1, 0, 1])
        hits = Table(
            {
                "channel_id": channel_ids,
                "dom_id": dom_ids,
                "time": times,
                "tot": tots,
                "triggered": triggereds,
                "group_id": 0,  # event_id
            },
            name="hits",
            h5loc="/foo",
            split_h5=True,
        )
        triggered_hits = hits.triggered_rows
        assert len(triggered_hits) == 3
        assert triggered_hits.split_h5
        assert triggered_hits.name == "hits"
        assert triggered_hits.h5loc == "/foo"

    def test_triggered_missing_col_raises(self):
        n = 5
        channel_ids = np.arange(n)
        dom_ids = np.arange(n)
        times = np.arange(n)
        tots = np.arange(n)
        hits = Table(
            {
                "channel_id": channel_ids,
                "dom_id": dom_ids,
                "time": times,
                "tot": tots,
                "group_id": 0,  # event_id
            },
            name="hits",
            h5loc="/foo",
            split_h5=True,
        )
        with pytest.raises(KeyError):
            triggered_hits = hits.triggered_rows
            assert triggered_hits is not None


class TestNDArray(TestCase):
    def test_init(self):
        arr = np.random.random((2, 3, 4))
        ndarr = NDArray(arr)

    def test_init_array(self):
        arr = np.random.random((2, 3, 4))
        arr = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        ndarr = NDArray(arr)
        assert 1 == ndarr[0, 0, 0]
        assert 6 == ndarr[1, 0, 1]
        assert 7 == ndarr[1, 1, 0]
        assert "/misc" == ndarr.h5loc
        assert "Unnamed NDArray" == ndarr.title
        assert ndarr.group_id is None

    def test_attributes(self):
        ndarr = NDArray([1], h5loc="/foo", title="Foo", group_id=23)
        assert "/foo" == ndarr.h5loc
        assert "Foo" == ndarr.title
        assert 23 == ndarr.group_id

    def test_slicing_preserves_attribute(self):
        ndarr = NDArray([1, 2, 3], h5loc="/foo", title="Foo", group_id=23)
        a = ndarr[:1]
        assert "/foo" == a.h5loc
        assert "Foo" == a.title
        assert 23 == a.group_id


class TestVec3(TestCase):
    def test_init(self):
        Vec3(0, 1, 2)

    def test_indices(self):
        v = Vec3(0, 1, 2)
        assert 0 == v[0]
        assert 1 == v[1]
        assert 2 == v[2]

    def test_attributes(self):
        v = Vec3(0, 1, 2)
        assert 0 == v.x
        assert 1 == v.y
        assert 2 == v.z

    def test_mutability(self):
        v = Vec3(0, 1, 2)
        v.x = 42
        assert 42 == v.x

    def test_vector_addition(self):
        u = Vec3(1, 2, 3)
        v = Vec3(4, 5, 6)
        w = u + v
        assert 5 == w.x
        assert 7 == w.y
        assert 9 == w.z

    def test_vector_subtraction(self):
        u = Vec3(1, 2, 3)
        v = Vec3(4, 7, 11)
        w = u - v
        assert -3 == w.x
        assert -5 == w.y
        assert -8 == w.z

    def test_scalar_addition(self):
        u = Vec3(1, 2, 3)

        v = u + 1
        assert 2 == v.x
        assert 3 == v.y
        assert 4 == v.z
        w = 1 + u
        assert 2 == w.x
        assert 3 == w.y
        assert 4 == w.z

    def test_scalar_subtraction(self):
        u = Vec3(1, 2, 3)

        v = u - 1
        assert 0 == v.x
        assert 1 == v.y
        assert 2 == v.z
        w = 1 - u
        assert 0 == w.x
        assert -1 == w.y
        assert -2 == w.z

    def test_multiplication(self):
        u = Vec3(1, 2, 3)
        v = u * 2
        assert 2 == v.x
        assert 4 == v.y
        assert 6 == v.z
        w = 2 * u
        assert 2 == w.x
        assert 4 == w.y
        assert 6 == w.z

    def test_division(self):
        u = Vec3(2, 4, 8)
        v = u / 2
        assert 1 == v.x
        assert 2 == v.y
        assert 4 == v.z

    def test_linalg_norm(self):
        u = Vec3(1, 0, 0)
        norm = np.linalg.norm(u)
        assert 1 == norm
        u = Vec3(0, 3, 0)
        norm = np.linalg.norm(u)
        assert 3 == norm
        u = Vec3(1, 1, 0)
        norm = np.linalg.norm(u)
        assert np.sqrt(2) == norm



# -*- coding: utf-8 -*-
"""

==================
ToT histogram.
==================

Create a simple histogram of the PMT signals (ToTs) in all events.
"""

# Author: Tamas Gal <tgal@km3net.de>
# License: BSD-3

import tables as tb
import matplotlib.pyplot as plt
import km3pipe.style
from km3net_testdata import data_path

km3pipe.style.use("km3pipe")

filename = data_path("hdf5/atmospheric_muons_sample.h5")

with tb.File(filename) as f:
    tots = f.get_node("/hits/tot")[:]


fig, ax = plt.subplots()
ax.hist(tots, bins=max(tots), log=True, edgecolor="none")
ax.set_title("ToT distribution")
ax.set_xlabel("ToT [ns]")

fig.tight_layout()
plt.show()



try:
    import urwid
except ImportError:
    print("Could not import the Python module 'urwid'.")
    raise SystemExit


def handle_input(input):
    if input in UI.keys["escape"]:
        raise urwid.ExitMainLoop()
    if input in UI.keys["left"] + UI.keys["right"]:
        main_frame.set_focus("footer")
        next_blob()
    if input in UI.keys["down"] + UI.keys["up"]:
        main_frame.set_focus("body")


class UI(object):
    """Represents the settings for the UI."""

    fg = "light gray"
    bg = "black"

    palette = [
        ("default", fg, bg),
        ("highlight", fg + ",standout", bg),
        ("header", "white", "dark cyan"),
        ("footer", "light gray", "dark blue"),
        ("body", "dark cyan", "", "standout"),
        ("focus", "dark red", "", "standout"),
        ("head", "light red", "black"),
        ("blob", "yellow", "dark cyan"),
        ("blob_scale", "dark cyan", "black"),
    ]

    keys = {
        "select": ("return", "enter"),
        "inspect": ("x", "X"),
        "escape": ("esc", "q", "Q"),
        "left": ("left", "h"),
        "right": ("right", "l"),
        "up": ("up", "k"),
        "down": ("down", "j"),
        "goto": ("g", "G"),
        "help": ("?",),
    }


class ItemWidget(urwid.WidgetWrap):
    def __init__(self, id, description):
        self.id = id
        self.content = "item %s: %s..." % (str(id), description[:25])
        self.item = [
            (
                "fixed",
                15,
                urwid.Padding(
                    urwid.AttrWrap(urwid.Text("item %s" % str(id)), "body", "focus"),
                    left=2,
                ),
            ),
            urwid.AttrWrap(urwid.Text("%s" % description), "body", "focus"),
        ]
        w = urwid.Columns(self.item)
        self.__super.__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


class BlobSelector(urwid.WidgetWrap):
    def __init__(self, description):
        self.content = description
        self.item = [
            urwid.AttrWrap(urwid.Text("%s" % description), "blob", "focus"),
        ]
        w = urwid.Columns(self.item)
        self.__super.__init__(w)

    def selectable(self):
        return True

    def keypress(self, size, key):
        return key


class BlobWidget(urwid.Pile):
    def __init__(self):
        self.width = 20
        self.size = (0,)
        urwid.Pile.__init__(
            self,
            [
                urwid.Text("", wrap="clip"),
                urwid.Text("", wrap="clip"),
                urwid.Text("", wrap="clip"),
            ],
        )

    def draw(self):
        self.widget_list[0].set_text(".OOOOOOOOOOOOOOOOOOOO")
        self.widget_list[1].set_text("|    '    |    '    |")
        self.widget_list[2].set_text("0         10        20")

    def render(self, size, focus):
        self.size = size
        self.draw()
        return urwid.Pile.render(self, size, focus)


def next_blob():
    pass


header = urwid.AttrWrap(urwid.Text("The header!", align="center"), "header")
footer = urwid.AttrWrap(urwid.Text("The footer"), "footer")

items = []
for i in range(100):
    items.append(ItemWidget(i, "Item {0}".format(i)))

browser_header = urwid.AttrMap(urwid.Text("selected:"), "head")
browser_listbox = urwid.ListBox(urwid.SimpleListWalker(items))
browser_view = urwid.Frame(
    urwid.AttrWrap(browser_listbox, "body"), header=browser_header
)

blobs = BlobWidget()
footer = urwid.Columns([urwid.Text("Info\nSecond kube"), blobs])

main_frame = urwid.AttrWrap(urwid.Frame(browser_view, focus_part="body"), "default")
main_frame.set_header(header)
main_frame.set_footer(footer)

loop = urwid.MainLoop(main_frame, UI.palette, unhandled_input=handle_input)
loop.run()



# Filename: test_time.py
# pylint: disable=locally-disabled,C0111,R0904,C0103

from datetime import datetime, timedelta
from time import sleep

from km3pipe.testing import TestCase, MagicMock
from km3pipe.time import Cuckoo, total_seconds, Timer

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class TestTools(TestCase):
    def test_total_seconds(self):
        seconds = 3
        time1 = datetime.now()
        time2 = time1 + timedelta(seconds=seconds)
        td = time2 - time1
        self.assertAlmostEqual(seconds, total_seconds(td))


class TestCuckoo(TestCase):
    def test_reset_timestamp(self):
        cuckoo = Cuckoo()
        cuckoo.reset()
        delta = datetime.now() - cuckoo.timestamp
        self.assertGreater(total_seconds(delta), 0)

    def test_set_interval_on_init(self):
        cuckoo = Cuckoo(1)
        self.assertEqual(1, cuckoo.interval)

    def test_set_callback(self):
        callback = 1
        cuckoo = Cuckoo(callback=callback)
        self.assertEqual(1, cuckoo.callback)

    def test_msg_calls_callback(self):
        callback = MagicMock()
        message = "a"
        cuckoo = Cuckoo(callback=callback)
        cuckoo.msg(message)
        callback.assert_called_with(message)

    def test_msg_calls_callback_with_empty_args(self):
        callback = MagicMock()
        cuckoo = Cuckoo(callback=callback)
        cuckoo.msg()
        callback.assert_called_with()

    def test_msg_calls_callback_with_multiple_args(self):
        callback = MagicMock()
        cuckoo = Cuckoo(callback=callback)
        cuckoo.msg(1, 2, 3)
        callback.assert_called_with(1, 2, 3)

    def test_msg_calls_callback_with_multiple_kwargs(self):
        callback = MagicMock()
        cuckoo = Cuckoo(callback=callback)
        cuckoo.msg(a=1, b=2)
        callback.assert_called_with(a=1, b=2)

    def test_msg_calls_callback_with_mixed_args_and_kwargs(self):
        callback = MagicMock()
        cuckoo = Cuckoo(callback=callback)
        cuckoo.msg(1, 2, c=3, d=4)
        callback.assert_called_with(1, 2, c=3, d=4)

    def test_direct_call_calls_callback(self):
        callback = MagicMock()
        message = "a"
        cuckoo = Cuckoo(callback=callback)
        cuckoo(message)
        callback.assert_called_with(message)

    def test_msg_is_not_called_when_interval_not_reached(self):
        callback = MagicMock()
        message = "a"
        cuckoo = Cuckoo(10, callback)
        cuckoo.reset()
        cuckoo.msg(message)
        self.assertFalse(callback.called)

    def test_msg_is_only_called_when_interval_reached(self):
        callback = MagicMock()
        message = "a"
        cuckoo = Cuckoo(0.1, callback)
        cuckoo.reset()
        cuckoo.msg(message)
        self.assertFalse(callback.called)
        sleep(0.11)
        cuckoo.msg(message)
        self.assertTrue(callback.called)

    def test_msg_sets_timestamp_on_first_call(self):
        cuckoo = Cuckoo()
        cuckoo.msg()
        assert cuckoo.timestamp

    def test_msg_gets_called_on_the_very_first_time(self):
        callback = MagicMock()
        message = "a"
        cuckoo = Cuckoo(1, callback)
        cuckoo.msg(message)
        self.assertTrue(callback.called)

    def test_msg_resets_timestamp_after_interval_reached(self):
        callback = MagicMock()
        message = "a"
        cuckoo = Cuckoo(0.1, callback)
        cuckoo.reset()
        timestamp1 = cuckoo.timestamp
        print(cuckoo.timestamp)
        self.assertFalse(callback.called)
        sleep(0.11)
        cuckoo.msg(message)
        timestamp2 = cuckoo.timestamp
        print(cuckoo.timestamp)
        self.assertTrue(callback.called)
        assert timestamp1 is not timestamp2

    def test_interval_reached(self):
        cuckoo = Cuckoo(0.1)
        cuckoo.reset()
        self.assertFalse(cuckoo._interval_reached())
        sleep(0.11)
        self.assertTrue(cuckoo._interval_reached())


class TestTimer(TestCase):
    def test_context_manager(self):
        mock = MagicMock()
        with Timer(callback=mock) as t:  # noqa
            pass
        mock.assert_called_once()

    def test_context_manager_calls_with_standard_text(self):
        mock = MagicMock()
        with Timer(callback=mock) as t:  # noqa
            pass
        self.assertTrue(mock.call_args[0][0].startswith("It "))



# Filename: logger.py
# pylint: disable=locally-disabled,C0103
"""
The logging facility.

"""
from hashlib import sha256
from inspect import getframeinfo, stack
import socket
import logging
import logging.handlers

from .tools import colored, supports_color

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

loggers = {}  # this holds all the registered loggers
# logging.basicConfig()

DEPRECATION = 45
logging.addLevelName(DEPRECATION, "DEPRECATION")
ONCE = 46
logging.addLevelName(ONCE, "ONCE")


def deprecation(self, message, *args, **kws):
    """Show a deprecation warning."""
    self._log(DEPRECATION, message, args, **kws)


def once(self, message, *args, **kws):
    """Show a message only once, determined by position in source or identifer.

    This will not work in IPython or Jupyter notebooks if no identifier is
    specified, since then the determined position in source contains the
    execution number of the input (cell), which changes every time.
    Set a unique ``identifier=X``, otherwise the message will be printed every
    time.

    """
    identifier = kws.pop("identifier", None)

    if identifier is None:
        caller = getframeinfo(stack()[1][0])
        identifier = "%s:%d" % (caller.filename, caller.lineno)
    if not hasattr(self, "once_dict"):
        self.once_dict = {}
    if identifier in self.once_dict:
        return
    self.once_dict[identifier] = True
    self._log(ONCE, message, args, **kws)


logging.Logger.deprecation = deprecation
logging.Logger.once = once

if supports_color():
    logging.addLevelName(
        logging.INFO, "\033[1;32m%s\033[1;0m" % logging.getLevelName(logging.INFO)
    )
    logging.addLevelName(
        logging.DEBUG, "\033[1;34m%s\033[1;0m" % logging.getLevelName(logging.DEBUG)
    )
    logging.addLevelName(
        logging.WARNING, "\033[1;33m%s\033[1;0m" % logging.getLevelName(logging.WARNING)
    )
    logging.addLevelName(
        logging.ERROR, "\033[1;31m%s\033[1;0m" % logging.getLevelName(logging.ERROR)
    )
    logging.addLevelName(
        logging.CRITICAL,
        "\033[1;101m%s\033[1;0m" % logging.getLevelName(logging.CRITICAL),
    )
    logging.addLevelName(DEPRECATION, "\033[1;35m%s\033[1;0m" % "DEPRECATION")
    logging.addLevelName(ONCE, "\033[1;36m%s\033[1;0m" % "ONCE")


class LogIO(object):
    """Read/write logging information."""

    def __init__(self, node, stream, url="pi2089.physik.uni-erlangen.de", port=28777):
        self.node = node
        self.stream = stream
        self.url = url
        self.port = port
        self.sock = None
        self.connect()

    def send(self, message, level="info"):
        message_string = "+log|{0}|{1}|{2}|{3}\r\n".format(
            self.stream, self.node, level, message
        )
        try:
            self.sock.send(message_string)
        except socket.error:
            print("Lost connection, reconnecting...")
            self.connect()
            self.sock.send(message_string)

    def connect(self):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.connect((self.url, self.port))


def get_logger(name, filename=None, stream_loglevel="INFO", file_loglevel="DEBUG"):
    """Helper function to get a logger"""
    if name in loggers:
        return loggers[name]
    logger = logging.getLogger(name)
    logger.propagate = False

    with_color = supports_color()

    pre1, suf1 = hash_coloured_escapes(name) if with_color else ("", "")
    pre2, suf2 = hash_coloured_escapes(name + "salt") if with_color else ("", "")
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)s {}+{}+{} "
        "%(name)s: %(message)s".format(pre1, pre2, suf1),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    if filename is not None:
        ch_file = logging.handlers.RotatingFileHandler(
            filename, maxBytes=5 * 1024 * 1024, backupCount=10
        )
        ch_file.setLevel(file_loglevel)
        ch_file.setFormatter(formatter)
        logger.addHandler(ch_file)
    ch = logging.StreamHandler()
    ch.setLevel(stream_loglevel)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    loggers[name] = logger

    logger.once_dict = {}

    return logger


def available_loggers():
    """Return a list of avialable logger names"""
    return list(logging.Logger.manager.loggerDict.keys())


def set_level(log_or_name, level):
    """Set the log level for given logger and all handlers"""
    if isinstance(log_or_name, str):
        log = get_logger(log_or_name)
    else:
        log = log_or_name
    log.setLevel(level)
    for handler in log.handlers:
        handler.setLevel(level)


def get_printer(name, color=None, ansi_code=None, force_color=False):
    """Return a function which prints a message with a coloured name prefix"""

    if force_color or supports_color():
        if color is None and ansi_code is None:
            cpre_1, csuf_1 = hash_coloured_escapes(name)
            cpre_2, csuf_2 = hash_coloured_escapes(name + "salt")
            name = cpre_1 + "+" + cpre_2 + "+" + csuf_1 + " " + name
        else:
            name = colored(name, color=color, ansi_code=ansi_code)

    prefix = name + ": "

    def printer(text):
        print(prefix + str(text))

    return printer


def hash_coloured(text):
    """Return a ANSI coloured text based on its hash"""
    ansi_code = int(sha256(text.encode("utf-8")).hexdigest(), 16) % 230
    return colored(text, ansi_code=ansi_code)


def hash_coloured_escapes(text):
    """Return the ANSI hash colour prefix and suffix for a given text"""
    ansi_code = int(sha256(text.encode("utf-8")).hexdigest(), 16) % 230
    prefix, suffix = colored("SPLIT", ansi_code=ansi_code).split("SPLIT")
    return prefix, suffix



# Filename: __init__.py
# -*- coding: utf-8 -*-
# pylint: disable=locally-disabled
"""
A collection of commonly used modules.

"""

from .common import Dump, Keep, Delete, StatusBar
from .mc import GlobalRandomState

from . import ahrs
from . import common
from . import communication
from . import fit
from . import hardware
from . import hits
from . import io
from . import k40
from . import mc
from . import parser
from . import plot
from . import topology
